
# $Id: UCD.pm,v 1.6 2000/06/12 15:28:21 lth Exp $

=head1 NAME

CHIRP::UCD::UCD - a CHIRP driver for hosts running the UCD-SNMP
agent.

=head1 DESCRIPTION

CHIRP::UCD::UCD is a driver for the CHIRP utility, that generates
Cricket configuration files.

The driver is specifically written to extract as much useful
information as possible from hosts that are running the UCD-SNMP
agent.

The driver tries very hard to be flexible and intelligent about which
parts of the MIB that is supported in the particular instance of the
UCD-SNMP agent. For instance, if the HOST-RESOURCES-MIB is compiled
into the agent, this is detected by the driver, and additional targets
are output to the resulting configuration file.

The resulting configuration file will include the following sections,
depending on the support in the MIB:

=head2 hr_System - number of system processes and users

This target is only generated, when the HOST-RESOURCES-MIB is included
in the ucd-snmp agent.

This section is not fully implemented yet.

=head2 ucd_System - CPU, Memory and Load

This target lists the CPU utilisation, Memory utilisation and Average
Load of the host.

On the CPU page, the user and system utilisation is shown.

On the Memory page, the amount of swap space is shown, together with
the available and used RAM.

On the Load page, the 5 and 15 minutes load averages are shown.

=head2 interface - octets, unicast packets and errors

This section lists each network interface, and includes the
input/output octets, unicast packets, and errors on each interface.

Each target in this section is named 'intf_' and then the interface
name, e.g. 'intf_fxp0'.

=head2 hr_Storage - Storage utilisation

The targets in this section are only generated, when the
HOST-RESOURCES-MIB is included in the ucd-snmp agent.

Each page show the size of the storage, and the amount that is
allocated.

Each target in this section is named 'stor_' and then the mount point,
e.g. 'stor_var'.

=head2 ucd_Storage - Storage utilisation

The targets in this section are only generated, when individual 'disk
PATH' entries have been entered into the agent's configuration file.

This section is not properly implemented yet.

=head1 INSTALLATION

Put the UCD.pm script in $CHIRP/lib/UCD/UCP.pm.

Add these lines to the $CHIRP/id.cfg file:

 Device  UCD_UCD
        Module  = "lib::UCD::UCD"
        Package = "CHIRP::UCD::UCD"
        Vendor  = "University of California, Davis"
        Model   = "Various"
        Match   = "OID0"
        OID0    = "snmp://%snmp%/sysObjectID=1.3.6.1.4.1.2021.250"

There is one hardcoded path in UCD.pm, that you might need to fix up: the path for the Cricket libraries in the 'use lib'-statement below.

=head1 SEE ALSO

Cricket contribs (including CHIRP): http://www.gnac.com/techinfo/cricket_contrib/index.html

Cricket: http://cricket.sourceforge.net

UCD-SNMP: http://ucd-snmp.ucdavis.edu

=head1 BUGS

Probably. Patches are welcome.

=head1 CREDITS

CHIRP is written by Tom Honermann <tom@honermann.net> and Scott
Feldmann <sm.feldmann@hosp.wisc.edu> for the University of Wisconsin
Hospital & Clinics.

Cricket is written by Jeff Allen <jra@corp.webtv.net> of WebTV.

=head1 COPYRIGHT

Copyright (c) 2000 Lars Thegler. All rights reserved.

Redistribution and use, with or without modification, are permitted
provided that the above copyright notice is retained.

=head1 AUTHOR

Lars Thegler <lars@thegler.dk>

Contributions from Scott Feldmann <sm.feldmann@hosp.wisc.edu>.

=cut

use strict;
$| = 1;

package CHIRP::UCD::UCD;

use lib "/usr/local/cricket/lib";
use SNMP_util ();

my $DEBUG = 0;  # set this to 1 to log all SNMP traffic

sub wrap {
    # ignore the warnings that SNMP_util may yield
    my($name, $func, @argv) = @_;
    my @res;
    my @err;
    {
	local($SIG{'__WARN__'}) = sub { @err = @_; };
	@res = &{$func}(@argv);
    }
    if ($DEBUG) {
	printf "#%s(%s) = (%s)%s\n",
	  $name,
	  join(', ', map {"'$_'"} @argv),
	  join(', ', map {defined $_ ? "'$_'" : "undef"} @res),
	  @err ? map {" [$_]"} join(', ', map {s/\n//; $_} @err) : "";
    }
    if (@err) {
	return;
    } else {
	return @res;
    }
}

sub get {
    wrap('get', \&SNMP_util::snmpget, @_);
}

sub walk {
    wrap('walk', \&SNMP_util::snmpwalk, @_);
}

sub GenConfig {
    my($hostname, $snmp_community, $snmp_port, $options) = @_;

    $snmp_community = 'public' unless defined($snmp_community);
    $snmp_port = '161' unless defined($snmp_port);

    my($snmp) = "$snmp_community\@$hostname:$snmp_port";
    my($config) = {};

    my $oid_sysName  = '.1.3.6.1.2.1.1.5.0';
    my $oid_hrSystem = '.1.3.6.1.2.1.25.1';

    ##
    # Get the system name
    #
    my ($sysName) = get($snmp, $oid_sysName);

    ##
    # Determine whether the Host Resources MIB is available
    #
    my $hr_available = 0;
    if (walk($snmp, $oid_hrSystem)) {
	$hr_available = 1;
    }

    ##
    #	OIDs
    #

    $config->{oid}->{ifInOctets}         = '.1.3.6.1.2.1.2.2.1.10';
    $config->{oid}->{ifInUcastPkts}      = '.1.3.6.1.2.1.2.2.1.11';
    $config->{oid}->{ifInErrors}         = '.1.3.6.1.2.1.2.2.1.14';
    $config->{oid}->{ifOutOctets}        = '.1.3.6.1.2.1.2.2.1.16';
    $config->{oid}->{ifOutUcastPkts}     = '.1.3.6.1.2.1.2.2.1.17';
    $config->{oid}->{ifOutErrors}        = '.1.3.6.1.2.1.2.2.1.20';

    if ($hr_available) {
	# Host Resources System Group
	$config->{oid}->{hrSystemNumUsers}   = '.1.3.6.1.2.1.25.1.5.0';
	$config->{oid}->{hrSystemProcesses}  = '.1.3.6.1.2.1.25.1.6';
        $config->{oid}->{hrStorageSize}      = '.1.3.6.1.2.1.25.2.3.1.5';
	$config->{oid}->{hrStorageUsed}      = '.1.3.6.1.2.1.25.2.3.1.6';
    }

    # Memory stats
    $config->{oid}->{ucd_memAvailSwap}   = '.1.3.6.1.4.1.2021.4.4.0';
    $config->{oid}->{ucd_memAvailReal}   = '.1.3.6.1.4.1.2021.4.6.0';
    $config->{oid}->{ucd_memTotalFree}   = '.1.3.6.1.4.1.2021.4.11.0';

    # Disk stats
    $config->{oid}->{ucd_dskTotal}       = '.1.3.6.1.4.1.2021.9.1.6';
    $config->{oid}->{ucd_dskAvail}       = '.1.3.6.1.4.1.2021.9.1.7';
    $config->{oid}->{ucd_dskUsed}        = '.1.3.6.1.4.1.2021.9.1.8';
    $config->{oid}->{ucd_dskPercent}     = '.1.3.6.1.4.1.2021.9.1.9';
    $config->{oid}->{ucd_dskPercentNode} = '.1.3.6.1.4.1.2021.9.1.10';

    # Load averages
    $config->{oid}->{ucd_load5min}       = '.1.3.6.1.4.1.2021.10.1.3.2';
    $config->{oid}->{ucd_load15min}      = '.1.3.6.1.4.1.2021.10.1.3.3';

    # CPU Stats
    $config->{oid}->{ucd_ssCpuUser}      = '.1.3.6.1.4.1.2021.11.9.0';
    $config->{oid}->{ucd_ssCpuSystem}    = '.1.3.6.1.4.1.2021.11.10.0';

    ##
    #	DataSources
    #

    if ($hr_available) {

	# Host Resources system stuff

	$config->{datasource}->{hrSystemNumUsers} =
	  {
	   'rrd-ds-type' => 'GAUGE',
	   'ds-source' => "snmp://%snmp%/hrSystemNumUsers",
	   'desc' => "The number of user sessions for which this host is storing state information. A session is a collection of processes requiring a single act of user authentication and possibly subject to collective job control.",
	  };

	$config->{datasource}->{hrSystemProcesses} =
	  {
	   'rrd-ds-type' => 'GAUGE',
	   'ds-source' => "snmp://%snmp%/hrSystemProcesses",
	   'desc'      => "The number of process contexts currently loaded or running on this system.",
	  };

	$config->{datasource}->{hrStorageSize} =
	  {
	   'rrd-ds-type' => 'GAUGE',
	   'ds-source' => "snmp://%snmp%/hrStorageSize.%inst%",
           'desc' => "The size of the storage represented by this entry.",
	  };

	$config->{datasource}->{hrStorageUsed} =
	  {
	   'rrd-ds-type' => 'GAUGE',
	   'ds-source' => "snmp://%snmp%/hrStorageUsed.%inst%",
           'desc' => "The amount of the storage represented by this entry that is allocated.",
	  };
    }

    # Memory stats

    $config->{datasource}->{ucd_memAvailSwap} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_memAvailSwap",
       'desc' => "Available Swap Space on the host.",
      };

    $config->{datasource}->{ucd_memAvailReal} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_memAvailReal",
       'desc' => "Available Real/Physical Memory Space on the host.",
      };

    $config->{datasource}->{ucd_memTotalFree} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_memTotalFree",
       'desc' => "Total Available Memory on the host.",
      };

    # Disk stats

    $config->{datasource}->{ucd_dskTotal} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_dskTotal.%inst%",
       'desc'      => "Total size of the disk/partion.",
      };

    $config->{datasource}->{ucd_dskAvail} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_dskAvail.%inst%",
       'desc'      => "Available space on the disk.",
      };

    $config->{datasource}->{ucd_dskUsed} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_dskUsed.%inst%",
       'desc'      => "Used space on the disk.",
      };

    $config->{datasource}->{ucd_dskPercent} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_dskPercent.%inst%",
       'desc'      => "Percentage of space used on disk.",
      };

    $config->{datasource}->{ucd_dskPercentNode} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_dskPercentNode.%inst%",
       'desc'      => "Percentage of inodes used on disk.",
      };

    # Load averages

    $config->{datasource}->{ucd_load5min} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_load5min",
       'desc' => "The 5 minute load average.",
      };

    $config->{datasource}->{ucd_load15min} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_load15min",
       'desc' => "The 15 minute load average.",
      };

    # CPU Stats

    $config->{datasource}->{ucd_ssCpuUser} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_ssCpuUser",
       'desc' => "percentages of user CPU time.",
      };

    $config->{datasource}->{ucd_ssCpuSystem} =
      {
       'rrd-ds-type' => 'GAUGE',
       'ds-source' => "snmp://%snmp%/ucd_ssCpuSystem",
       'desc' => "percentages of system CPU time.",
      };

    # Interfaces stats

    $config->{datasource}->{ifInOctets} =
      {
       'rrd-ds-type' => 'COUNTER',
       'ds-source' => 'snmp://%snmp%/ifInOctets.%inst%',
       'desc' => "The total number of octets received on the interface, including framing characters.",
      };

    $config->{datasource}->{ifOutOctets} =
      {
       'rrd-ds-type' => 'COUNTER',
       'ds-source' => 'snmp://%snmp%/ifOutOctets.%inst%',
       'desc' => "The total number of octets transmitted out of the interface, including framing characters.",
      };

    $config->{datasource}->{ifInUcastPkts} =
      {
       'rrd-ds-type' => 'COUNTER',
       'ds-source' => 'snmp://%snmp%/ifInUcastPkts.%inst%',
       'desc' => "The number of packets, delivered by this sub-layer to a higher (sub-)layer, which were not addressed to a multicast or broadcast address at this sub-layer.",
      };

    $config->{datasource}->{ifOutUcastPkts} =
      {
       'rrd-ds-type' => 'COUNTER',
       'ds-source' => 'snmp://%snmp%/ifOutUcastPkts.%inst%',
       'desc' => "The total number of packets that higher-level protocols requested be transmitted, and which were not addressed to a multicast or broadcast address at this sub-layer, including those that were discarded or not sent.",
      };

    $config->{datasource}->{ifInErrors} =
      {
       'rrd-ds-type' => 'COUNTER',
       'ds-source' => 'snmp://%snmp%/ifInErrors.%inst%',
       'desc' => 'For packet-oriented interfaces, the number of inbound packets that contained errors preventing them from being deliverable to a higher-layer protocol.  For character-oriented or fixed-length interfaces, the number of inbound transmission units that contained errors preventing them from being deliverable to a higher-layer protocol.',
      };

    $config->{datasource}->{ifOutErrors} =
      {
       'rrd-ds-type' => 'COUNTER',
       'ds-source' => 'snmp://%snmp%/ifOutErrors.%inst%',
       'desc' => "For packet-oriented interfaces, the number of outbound packets that could not be transmitted because of errors.  For character-oriented or fixed-length interfaces, the number of outbound transmission units that could not be transmitted because of errors.",
      };


    ##
    #	TargetTypes
    #

    if ($hr_available) {

	$config->{targettype}->{hr_System} =
	  {
	   'ds'   => "hrSystemProcesses, hrSystemNumUsers",
	   'view' => "Processes: hrSystemProcesses, Users: hrSystemNumUsers",
	  };

	$config->{targettype}->{hr_Storage} =
	  {
	   'ds'   => "hrStorageSize, hrStorageUsed",
	   'view' => "Storage: hrStorageSize hrStorageUsed",
	  };
    }

    $config->{targettype}->{ucd_System} =
      {
       'ds'   => "ucd_ssCpuUser, ucd_ssCpuSystem, ".
                 "ucd_memAvailSwap, ucd_memAvailReal, ucd_memTotalFree, ".
	         "ucd_load5min, ucd_load15min",
       'view' => "CPU: ucd_ssCpuUser ucd_ssCpuSystem, ".
                 "Memory: ucd_memAvailSwap ucd_memAvailReal ucd_memTotalFree, ".
                 "Load: ucd_load5min ucd_load15min",
      };

    $config->{targettype}->{ucd_Storage}=
      {
       'ds' => "ucd_dskTotal, ucd_dskAvail, ucd_dskUsed, ucd_dskPercent, ucd_dskPercentNode",
       'view' => "Storage: ucd_dskTotal ucd_dskAvail ucd_dskUsed, ".
                 "Percent: ucd_dskPercent ucd_dskPercentNode",
      };

    $config->{targettype}->{standard_interface} =
      {
       'ds' => "ifInOctets, ifOutOctets, ".
               "ifInErrors, ifOutErrors, ".
               "ifInUcastPkts, ifOutUcastPkts",
       'view' => "Octets: ifInOctets ifOutOctets, ".
                 "UcastPkts: ifInUcastPkts ifOutUcastPkts, ".
                 "Errors: ifInErrors ifOutErrors",
      };

    ##
    #	Graphs
    #

    $config->{graph}->{"--default--"}->{"y-max"} = 'undef';

    if ($hr_available) {

	$config->{graph}->{hrSystemNumUsers} = {
						'color'     => 'dark-green',
						'draw-as'   => 'AREA',
						'y-axis'    => 'Users',
						'units'     => 'Users',
						'legend'    => '# of Users',
						'precision' => 'integer',
					       };

	$config->{graph}->{hrSystemProcesses} = {
						 'color'     => 'dark-green',
						 'draw-as'   => 'AREA',
						 'y-axis'    => 'Processes',
						 'units'     => 'Processes',
						 'legend'    => '# of system processes',
						 'precision' => 'integer',
						};

	$config->{graph}->{hrStorageSize} = {
					     'color' => 'blue',
					     'y-axis'  => "Bytes used",
					     'units' => "Bytes",
					     'legend' => "Bytes of space available",
					     'bytes'   => 'true',
					     'scale'   => '%units%',
					     'y-min'   => '%min-size%',
					     'y-max'   => '%max-size%',
					    };

	$config->{graph}->{hrStorageUsed} = {
					     'color' => 'dark-green',
					     'draw-as' => 'AREA',
					     'y-axis'  => "Bytes used",
					     'units'   => "Bytes",
					     'legend'  => "Bytes of space used",
					     'bytes'   => 'true',
					     'scale'   => '%units%',
					     'y-min'   => '%min-size%',
					     'y-max'   => '%max-size%',
					    };
    }

    $config->{graph}->{ucd_memAvailSwap} = {
					    'legend'  => "Available swap space",
					    'scale'   => '1024,*',
					    'bytes'   => 'true',
					    'units'   => 'Bytes',
					   };

    $config->{graph}->{ucd_memAvailReal} = {
					    'legend'  => "Used RAM",
					    'scale'   => '1024,*',
					    'bytes'   => 'true',
					    'units'   => 'Bytes',
					   };

    $config->{graph}->{ucd_memTotalFree} = {
					    'legend'  => "Available RAM",
					    'scale'   => '1024,*',
					    'bytes'   => 'true',
					    'units'   => 'Bytes',
					   };

    $config->{graph}->{ucd_dskTotal} = {
					'color'   => 'blue',
					'y-axis'  => "Bytes",
					'legend'  => "Bytes Total",
					'bytes'   => 'true',
					'y-max'   => '%max-size%',
					'y-min'   => '0',
					'scale'   => '1024,*',
				       };

    $config->{graph}->{ucd_dskAvail} = {
					'color'   => 'dark-green',
					'draw-as' => 'AREA',
					'y-axis'  => "Bytes",
					'legend'  => "Bytes of space free",
					'bytes'   => 'true',
					'y-max'   => '%max-size%',
					'y-min'   => '0',
					'scale'   => '1024,*',
				       };

    $config->{graph}->{ucd_dskUsed} = {
					'color'   => 'red',
					'draw-as' => 'AREA',
					'y-axis'  => "Bytes",
					'legend'  => "Bytes of space used",
					'bytes'   => 'true',
					'y-max'   => '%max-size%',
					'y-min'   => '0',
					'scale'   => '1024,*',
				       };

    $config->{graph}->{ucd_dskPercent} = {
					 'color'   => 'blue',
					 'draw-as' => 'AREA',
					 'y-axis'  => "%",
					 'units'   => "%",
					 'legend'  => "Percent of space used",
					};

    $config->{graph}->{ucd_dskPercentNode} = {
					      'color'   => 'blue',
					      'draw-as' => 'AREA',
					      'y-axis'  => "%",
					      'units'   => "%",
					      'legend'  => "Percent of inodes used",
					     };

    $config->{graph}->{ucd_load5min} = {
					'legend'   => "5 minute load average",
					'si-units' => 'false',
				       };

    $config->{graph}->{ucd_load15min} = {
					 'legend'   => "15 minute load average",
					 'si-units' => 'false',
					};

    $config->{graph}->{ucd_ssCpuUser} = {
					 'legend' => "CPU user time",
					 'y-axis' => "Percent",
					 'units'  => "%",
					};

    $config->{graph}->{ucd_ssCpuSystem} = {
					 'legend' => "CPU system time",
					 'y-axis' => "Percent",
					 'units'  => "%",
					};

    $config->{graph}->{ifInOctets} = {
				      'color' => 'dark-green',
				      'draw-as' => 'AREA',
				      'legend' => "Average bits in",
				      'y-axis' => "bits per second",
				      'units' => "bits/sec",
				      'scale' => '8,*',
				      'bytes' => 'true',
				     };

    $config->{graph}->{ifOutOctets} = {
				       'color' => 'blue',
				       'legend' => "Average bits out",
				       'y-axis' => "bits per second",
				       'units' => "bits/sec",
				       'scale' => '8,*',
				       'bytes' => 'true',
				      };

    $config->{graph}->{ifInUcastPkts} = {
					 'color'   => 'dark-green',
					 'draw-as' => 'AREA',
					 'y-axis'  => "packets per second",
					 'units'   => "pkt/sec",
					 'legend'  => "Average num Packets In",
					};

    $config->{graph}->{ifOutUcastPkts} = {
					  'color'   => 'blue',
					  'y-axis'  => "packets per second",
					  'units'   => "pkt/sec",
					  'legend'  => "Average num Packets Out",
					};

    $config->{graph}->{ifInErrors} = {
				      'color'   => 'dark-green',
				      'draw-as' => 'AREA',
				      'y-axis'  => "errors per second",
				      'units'   => "err/sec",
				      'legend'  => "Average Errors In",
				     };

    $config->{graph}->{ifOutErrors} = {
				      'color'   => 'blue',
				      'y-axis'  => "errors per second",
				      'units'   => "err/sec",
				      'legend'  => "Average Errors Out",
				     };


    ##
    #	Targets
    #

    $config->{target}->{"--default--"}->{"snmp-host"} = "$hostname";
    $config->{target}->{"--default--"}->{"snmp-community"} = "$snmp_community";
    $config->{target}->{"--default--"}->{"snmp-port"} = "$snmp_port";
    $config->{target}->{"--default--"}->{"snmp"} = "%snmp-community%@%snmp-host%:%snmp-port%";
    $config->{target}->{"--default--"}->{"directory-desc"} = $sysName;
    $config->{target}->{"--default--"}->{"short-desc"} = '""';
    $config->{target}->{"--default--"}->{"long-desc"} = "%short-desc%";

    if ($hr_available) {

	$config->{target}->{hr_System} = {
				       'target-type' => 'hr_System',
				       'short-desc'  => "# of system processes and users",
				      };

    }

    $config->{target}->{ucd_System} = {
				    'target-type' => 'ucd_System',
				    'short-desc'  => "CPU, Memory, and Load",
				   };

    ##
    # Get interface mapping
    #
    my $oid_ifDescr        = '.1.3.6.1.2.1.2.2.1.2';
    my $oid_ifType         = '.1.3.6.1.2.1.2.2.1.3';
    my $oid_ifMtu          = '.1.3.6.1.2.1.2.2.1.4';
    my $oid_ifSpeed        = '.1.3.6.1.2.1.2.2.1.5';
    my $oid_ifPhysAddress  = '.1.3.6.1.2.1.2.2.1.6';
    my $oid_ifAdminStatus  = '.1.3.6.1.2.1.2.2.1.7';
    my $oid_ifOperStatus   = '.1.3.6.1.2.1.2.2.1.8';
    my $oid_ipAdEntIfIndex = '.1.3.6.1.2.1.4.20.1.2';

    # ifType mapping, according to IANAifType-MIB
    my @ifType;
    while (<DATA>) {
	s/--.*//;  # remove all comments
	if (/\s*(.*)\((\d+)\)/) {
	    $ifType[$2] = $1;
	}
    }
    my $interface = {};
    foreach my $row (walk($snmp, $oid_ifDescr)){
	my($index, $ifDescr) = split /:/, $row;
	$interface->{$index}->{ifDescr} = $ifDescr;
    }
    foreach my $index (keys %$interface) {
	for (get($snmp, "$oid_ifType.$index")) {
	    $interface->{$index}->{ifType} =
	      sprintf "%s(%d)",
		(defined $ifType[$_]) ? $ifType[$_] : 'Unknown',
		  $_;
	}
	($interface->{$index}->{ifMtu})   = get($snmp, "$oid_ifMtu.$index");
	($interface->{$index}->{ifSpeed}) = get($snmp, "$oid_ifSpeed.$index");
	if (my @addr = get($snmp, "$oid_ifPhysAddress.$index")) {
	    $interface->{$index}->{ifPhysAddress} =
	      join (':',
		    map { sprintf "%02x", $_ }
		    unpack("C*", $addr[0])
		   );
	}
    }
    foreach my $row (walk($snmp, $oid_ipAdEntIfIndex)) {
	print "#ipAdEntIfIndex: $row\n" if $DEBUG;
	my($ipaddr, $index) = split /:/, $row;
	$interface->{$index}->{ipAddr} = $ipaddr;
    }

    if ($DEBUG) {
	use Data::Dumper;
	print map {"#$_\n"} split /\n/, Data::Dumper->Dump([$interface], ['*interface']);
    }

    foreach my $index (keys %$interface) {

	# easy way to get rid of boring 'virtual' interfaces
	next unless $interface->{$index}->{ifSpeed};

	my $shortdesc = $interface->{$index}->{ifDescr};
	my $desc;

	$desc .= "<tr><td>ifType</td><td>$interface->{$index}->{ifType}</td></tr>";
	$desc .= "<tr><td>ifMtu</td><td>$interface->{$index}->{ifMtu} octets</td></tr>";

	{
	    my($value, $prefix) = si_unit($interface->{$index}->{ifSpeed});
	    $desc .= "<tr><td>ifSpeed</td><td>${value} ${prefix}bits per second</td></tr>";
	}

	if ($interface->{$index}->{ifPhysAddress}) {
	    $desc .= "<tr><td>ifPhysAddress</td><td>$interface->{$index}->{ifPhysAddress}</td></tr>";
	}

	if ($interface->{$index}->{ipAddr}) {
	    $desc .= "<tr><td>ipAddr</td><td>$interface->{$index}->{ipAddr}</td></tr>";
	    $shortdesc .= " $interface->{$index}->{ipAddr}";
	}

	$config->{target}->{'intf_' . $interface->{$index}->{ifDescr}} =
	  {
	   'target-type' => 'standard_interface',
	   'inst' => "map(interface-name)",
	   'interface-name' => $interface->{$index}->{ifDescr},
	   'short-desc' => $shortdesc,
	   'long-desc' => "<table>$desc</table>",
	  };
    }


    ##
    # Get storage mapping
    #
    my $oid_hrStorageType               = '.1.3.6.1.2.1.25.2.3.1.2';
    my $oid_hrStorageDescr              = '.1.3.6.1.2.1.25.2.3.1.3';
    my $oid_hrStorageAllocationUnits    = '.1.3.6.1.2.1.25.2.3.1.4';
    my $oid_hrStorageSize               = '.1.3.6.1.2.1.25.2.3.1.5';
    my $oid_hrStorageUsed               = '.1.3.6.1.2.1.25.2.3.1.6';
    my $oid_hrFSRemoteMountPoint        = '.1.3.6.1.2.1.25.3.8.1.3';
    my $oid_hrFSType                    = '.1.3.6.1.2.1.25.3.8.1.4';
    my $oid_hrFSAccess                  = '.1.3.6.1.2.1.25.3.8.1.5';
    my $oid_hrFSBootable                = '.1.3.6.1.2.1.25.3.8.1.6';
    my $oid_hrFSStorageIndex            = '.1.3.6.1.2.1.25.3.8.1.7';

    my $storage = {};
    foreach my $row (walk($snmp, $oid_hrStorageType)){
	my($inst, $hrStorageType) = split /:/, $row;
	$storage->{$inst}->{hrStorageType} = $hrStorageType;
    }
    foreach my $row (walk($snmp, $oid_hrFSStorageIndex)) {
	my($inst, $hrFSStorageIndex) = split /:/, $row;
	$storage->{$hrFSStorageIndex}->{hrFSIndex} = $inst;
    }
    foreach my $inst (keys %$storage) {
	($storage->{$inst}->{hrStorageDescr}) = get($snmp, "$oid_hrStorageDescr.$inst");
	($storage->{$inst}->{hrStorageAllocationUnits}) = get($snmp, "$oid_hrStorageAllocationUnits.$inst");
	($storage->{$inst}->{hrStorageSize}) = get($snmp, "$oid_hrStorageSize.$inst");
	($storage->{$inst}->{hrStorageUsed}) = get($snmp, "$oid_hrStorageUsed.$inst");

	if (exists $storage->{$inst}->{hrFSIndex}) {
	    my $fsinst = $storage->{$inst}->{hrFSIndex};
	    ($storage->{$inst}->{hrFSRemoteMountPoint}) = get($snmp, "$oid_hrFSRemoteMountPoint.$fsinst");
	    ($storage->{$inst}->{hrFSType}) = get($snmp, "$oid_hrFSType.$fsinst");
	    ($storage->{$inst}->{hrFSAccess}) = get($snmp, "$oid_hrFSAccess.$fsinst");
	    ($storage->{$inst}->{hrFSBootable}) = get($snmp, "$oid_hrFSBootable.$fsinst");
	}
    }

    if ($DEBUG) {
	use Data::Dumper;
	print map {"#$_\n"} split /\n/, Data::Dumper->Dump([$storage], ['*storage']);
    }

    foreach my $inst (keys %$storage) {
	# remote filesystems are monitored on the file server, not here
	next if $storage->{$inst}->{hrFSRemoteMountPoint};

	$storage->{$inst}->{hrStorageDescr} =~ s|"||g;
	my $target = $storage->{$inst}->{hrStorageDescr};
	if ($target eq '/') { $target = 'root' }
	for ($target) {
	    s|^/||;  # strip leading slash
	    s|\s+|_|g;  # replace whitespace with an underscore
	    s|/|_|g;  # replace slash with an underscore
	}

	my $maxsize = $storage->{$inst}->{hrStorageAllocationUnits} * $storage->{$inst}->{hrStorageSize};

	my $desc = '';
	if ($storage->{$inst}->{hrStorageType}) {
	    $desc .= sprintf "<tr><td>hrStorageType</td><td>%s</td></tr>",
	      hrStorageTypes($storage->{$inst}->{hrStorageType});
	}
	if ($storage->{$inst}->{hrFSType}) {
	    $desc .= sprintf "<tr><td>hrFSType</td><td>%s</td></tr>",
	      hrFSTypes($storage->{$inst}->{hrFSType});
	}
	if ($storage->{$inst}->{hrFSAccess}) {
	    $desc .= sprintf "<tr><td>hrFSAccess</td><td>%s</td></tr>",
	      { '1' => 'readWrite(1)',
		'2' => 'readOnly(2)',
	      }->{$storage->{$inst}->{hrFSAccess}};
	}
	if ($storage->{$inst}->{hrFSBootable}) {
	    $desc .= sprintf "<tr><td>hrFSBootable</td><td>%s</td></tr>",
	      { '1' => 'true(1)',
		'2' => 'false(2)',
	      }->{$storage->{$inst}->{hrFSBootable}};
	}

	$config->{target}->{'stor_' . $target} =
	  {
	   'target-type' => 'hr_Storage',
	   'inst' => $inst,
	   'short-desc' => "Bytes used on $storage->{$inst}->{hrStorageDescr}",
	   'long-desc' => "<table>$desc</table>",
	   'max-size' => $maxsize,
	   'storage' => $target,
	   'units' => "$storage->{$inst}->{hrStorageAllocationUnits},*",
	   'bytes' => 'true',
	  };

    }

    return $config;
}

1;


sub si_unit {
    my($value, $bytes) = @_;
    return ($value, '') if ($value == 0);

    my(@symbol) = ('a', 'f', 'p', 'n', '&micro;', 'milli',
		   '',
		   'k', 'M', 'G', 'T', 'P', 'E');
    my($symbcenter) = 6;
    my($digits) = int(log(abs($value))/log(10) / 3);

    my($magfact);
    if ($bytes) {
	$magfact = 2 ** ($digits * 10);
    } else {
	$magfact = 10 ** ($digits * 3.0);
    }

    if ((($digits + $symbcenter) > 0) &&
	(($digits + $symbcenter) <= $#symbol)) {
	return ($value/$magfact, $symbol[$digits + $symbcenter]);
    } else {
	return ($value, '');
    }
}


sub hrStorageTypes {
    my $oid = shift;
    return "bad oid" unless $oid =~ m/^1\.3\.6\.1\.2\.1\.25\.2\.1\.(\d+)$/;
    my $index = $1;
    my $str =
      {
       1 => 'Other',
       2 => 'Ram',
       3 => 'VirtualMemory',
       4 => 'FixedDisk',
       5 => 'RemovableDisk',
       6 => 'FloppyDisk',
       7 => 'CompactDisc',
       8 => 'RamDisk',
       9 => 'FlashMemory',
       10 => 'NetworkDisk',
      }->{$index} || "unknown";
    return $str . "($index)";
}


sub hrFSTypes {
    my $oid = shift;
    return "bad oid" unless $oid =~ m/^1\.3\.6\.1\.2\.1\.25\.3\.9\.(\d+)$/;
    my $index = $1;
    my $str =
      {
       1 => 'Other',
       2 => 'Unknown',
       3 => 'BerkeleyFFS',
       4 => 'Sys5FS',
       5 => 'Fat',
       6 => 'HPFS',
       7 => 'HFS',
       8 => 'MFS',
       9 => 'NTFS',
       10 => 'VNode',
       11 => 'Journaled',
       12 => 'iso9660',
       13 => 'RockRidge',
       14 => 'NFS',
       15 => 'Netware',
       16 => 'AFS',
       17 => 'DFS',
       18 => 'Appleshare',
       19 => 'RFS',
       20 => 'DGCFS',
       21 => 'BFS',
       22 => 'FAT32',
       23 => 'LinuxExt2',
      }->{$index} || "unknown";
    return $str . "($index)";
}


# ifType mapping, from IANAifType-MIB, August 31, 1999

__DATA__
                   other(1),          -- none of the following
                   regular1822(2),
                   hdh1822(3),
                   ddnX25(4),
                   rfc877x25(5),
                   ethernetCsmacd(6),
                   iso88023Csmacd(7),
                   iso88024TokenBus(8),
                   iso88025TokenRing(9),
                   iso88026Man(10),
                   starLan(11),
                   proteon10Mbit(12),
                   proteon80Mbit(13),
                   hyperchannel(14),
                   fddi(15),
                   lapb(16),
                   sdlc(17),
                   ds1(18),            -- DS1-MIB
                   e1(19),             -- Obsolete see DS1-MIB
                   basicISDN(20),
                   primaryISDN(21),
                   propPointToPointSerial(22), -- proprietary serial
                   ppp(23),
                   softwareLoopback(24),
                   eon(25),            -- CLNP over IP 
                   ethernet3Mbit(26),
                   nsip(27),           -- XNS over IP
                   slip(28),           -- generic SLIP
                   ultra(29),          -- ULTRA technologies
                   ds3(30),            -- DS3-MIB
                   sip(31),            -- SMDS, coffee
                   frameRelay(32),     -- DTE only. 
                   rs232(33),
                   para(34),           -- parallel-port
                   arcnet(35),         -- arcnet
                   arcnetPlus(36),     -- arcnet plus
                   atm(37),            -- ATM cells
                   miox25(38),
                   sonet(39),          -- SONET or SDH 
                   x25ple(40),
                   iso88022llc(41),
                   localTalk(42),
                   smdsDxi(43),
                   frameRelayService(44),  -- FRNETSERV-MIB
                   v35(45),
                   hssi(46),
                   hippi(47),
                   modem(48),          -- Generic modem
                   aal5(49),           -- AAL5 over ATM
                   sonetPath(50),
                   sonetVT(51),
                   smdsIcip(52),       -- SMDS InterCarrier Interface
                   propVirtual(53),    -- proprietary virtual/internal
                   propMultiplexor(54),-- proprietary multiplexing
                   ieee80212(55),      -- 100BaseVG
                   fibreChannel(56),   -- Fibre Channel
                   hippiInterface(57), -- HIPPI interfaces     
                   frameRelayInterconnect(58), -- Obsolete use either
                                       -- frameRelay(32) or 
                                       -- frameRelayService(44).
                   aflane8023(59),     -- ATM Emulated LAN for 802.3
                   aflane8025(60),     -- ATM Emulated LAN for 802.5
                   cctEmul(61),        -- ATM Emulated circuit          
                   fastEther(62),      -- Fast Ethernet (100BaseT)
                   isdn(63),           -- ISDN and X.25           
                   v11(64),            -- CCITT V.11/X.21             
                   v36(65),            -- CCITT V.36                  
                   g703at64k(66),      -- CCITT G703 at 64Kbps
                   g703at2mb(67),      -- Obsolete see DS1-MIB
                   qllc(68),           -- SNA QLLC                    
                   fastEtherFX(69),    -- Fast Ethernet (100BaseFX)   
                   channel(70),        -- channel                     
                   ieee80211(71),      -- radio spread spectrum       
                   ibm370parChan(72),  -- IBM System 360/370 OEMI Channel
                   escon(73),          -- IBM Enterprise Systems Connection
                   dlsw(74),           -- Data Link Switching
                   isdns(75),          -- ISDN S/T interface
                   isdnu(76),          -- ISDN U interface
                   lapd(77),           -- Link Access Protocol D
                   ipSwitch(78),       -- IP Switching Objects
                   rsrb(79),           -- Remote Source Route Bridging
                   atmLogical(80),     -- ATM Logical Port
                   ds0(81),            -- Digital Signal Level 0
                   ds0Bundle(82),      -- group of ds0s on the same ds1
                   bsc(83),            -- Bisynchronous Protocol
                   async(84),          -- Asynchronous Protocol
                   cnr(85),            -- Combat Net Radio
                   iso88025Dtr(86),    -- ISO 802.5r DTR
                   eplrs(87),          -- Ext Pos Loc Report Sys
                   arap(88),           -- Appletalk Remote Access Protocol
                   propCnls(89),       -- Proprietary Connectionless Protocol
                   hostPad(90),        -- CCITT-ITU X.29 PAD Protocol
                   termPad(91),        -- CCITT-ITU X.3 PAD Facility
                   frameRelayMPI(92),  -- Multiproto Interconnect over FR
                   x213(93),           -- CCITT-ITU X213
                   adsl(94),           -- Asymmetric Digital Subscriber Loop
                   radsl(95),          -- Rate-Adapt. Digital Subscriber Loop
                   sdsl(96),           -- Symmetric Digital Subscriber Loop
                   vdsl(97),           -- Very H-Speed Digital Subscrib. Loop
                   iso88025CRFPInt(98), -- ISO 802.5 CRFP
                   myrinet(99),        -- Myricom Myrinet
                   voiceEM(100),       -- voice recEive and transMit
                   voiceFXO(101),      -- voice Foreign Exchange Office
                   voiceFXS(102),      -- voice Foreign Exchange Station
                   voiceEncap(103),    -- voice encapsulation
                   voiceOverIp(104),   -- voice over IP encapsulation
                   atmDxi(105),        -- ATM DXI
                   atmFuni(106),       -- ATM FUNI
                   atmIma (107),       -- ATM IMA                  
                   pppMultilinkBundle(108), -- PPP Multilink Bundle
                   ipOverCdlc (109),   -- IBM ipOverCdlc
                   ipOverClaw (110),   -- IBM Common Link Access to Workstn
                   stackToStack (111), -- IBM stackToStack
                   virtualIpAddress (112), -- IBM VIPA
                   mpc (113),          -- IBM multi-protocol channel support
                   ipOverAtm (114),    -- IBM ipOverAtm
                   iso88025Fiber (115), -- ISO 802.5j Fiber Token Ring
                   tdlc (116),         -- IBM twinaxial data link control
                   gigabitEthernet (117), -- Gigabit Ethernet
                   hdlc (118),         -- HDLC
                   lapf (119),         -- LAP F
                   v37 (120),          -- V.37
                   x25mlp (121),       -- Multi-Link Protocol
                   x25huntGroup (122), -- X25 Hunt Group
                   trasnpHdlc (123),   -- Transp HDLC
                   interleave (124),   -- Interleave channel
                   fast (125),         -- Fast channel
                   ip (126),           -- IP (for APPN HPR in IP networks)
                   docsCableMaclayer (127),  -- CATV Mac Layer
                   docsCableDownstream (128), -- CATV Downstream interface
                   docsCableUpstream (129),  -- CATV Upstream interface
                   a12MppSwitch (130), -- Avalon Parallel Processor
                   tunnel (131),       -- Encapsulation interface
                   coffee (132),       -- coffee pot
                   ces (133),          -- Circuit Emulation Service
                   atmSubInterface (134), -- ATM Sub Interface
                   l2vlan (135),       -- Layer 2 Virtual LAN using 802.1Q
                   l3ipvlan (136),     -- Layer 3 Virtual LAN using IP
                   l3ipxvlan (137),    -- Layer 3 Virtual LAN using IPX
                   digitalPowerline (138), -- IP over Power Lines       
                   mediaMailOverIp (139), -- Multimedia Mail over IP
                   dtm (140),        -- Dynamic syncronous Transfer Mode
                   dcn (141),    -- Data Communications Network
                   ipForward (142),    -- IP Forwarding Interface
                   msdsl (143),       -- Multi-rate Symmetric DSL
                   ieee1394 (144), -- IEEE1394 High Performance Serial Bus
                   if-gsn (145),       --   HIPPI-6400 
                   dvbRccMacLayer (146), -- DVB-RCC MAC Layer
                   dvbRccDownstream (147),  -- DVB-RCC Downstream Channel
                   dvbRccUpstream (148),  -- DVB-RCC Upstream Channel
                   atmVirtual (149),   -- ATM Virtual Interface
                   mplsTunnel (150),   -- MPLS Tunnel Virtual Interface
                   srp (151),   -- Spatial Reuse Protocol       
                   voiceOverAtm (152),  -- Voice Over ATM
                   voiceOverFrameRelay (153),   -- Voice Over Frame Relay 
                   idsl (154),          -- Digital Subscriber Loop over ISDN
                   compositeLink (155),  -- Avici Composite Link Interface
                   ss7SigLink (156),     -- SS7 Signaling Link 
                   propWirelessP2P (157),  --  Prop. P2P wireless interface
                   frForward (158),    -- Frame Forward Interface
                   rfc1483 (159),       -- Multiprotocol over ATM AAL5
                   usb (160),           -- USB Interface
                   ieee8023adLag(161)  -- IEEE 802.3ad Link Aggregate
